@@ -15,11 +15,11 @@
 /******************** Constants ********************/
 /* HTTP timeouts, by operation, in milliseconds
  */
-#define DEVICE_HTTP_TIMEOUT_DEVCAPS     5000
-#define DEVICE_HTTP_TIMEOUT_PRECHECK    5000
+#define DEVICE_HTTP_TIMEOUT_DEVCAPS     20000
+#define DEVICE_HTTP_TIMEOUT_PRECHECK    20000
 #define DEVICE_HTTP_TIMEOUT_SCAN        30000
 #define DEVICE_HTTP_TIMEOUT_LOAD        -1
-#define DEVICE_HTTP_TIMEOUT_CHECK       5000
+#define DEVICE_HTTP_TIMEOUT_CHECK       20000
 #define DEVICE_HTTP_TIMEOUT_CLEANUP     30000
 #define DEVICE_HTTP_TIMEOUT_CANCEL      30000
 
@@ -215,6 +215,7 @@ device_new (zeroconf_devinfo *devinfo)
     log_debug(dev->log, "device created");
 
     dev->proto_ctx.log = dev->log;
+    dev->proto_ctx.devinfo = dev->devinfo;
     dev->proto_ctx.devcaps = &dev->opt.caps;
 
     devopt_init(&dev->opt);
@@ -609,7 +610,7 @@ device_scanner_capabilities_callback (void *ptr, http_query *q)
         goto DONE;
     }
 
-    devcaps_dump(dev->log, &dev->opt.caps);
+    devcaps_dump(dev->log, &dev->opt.caps, true);
     devopt_set_defaults(&dev->opt);
 
     /* Update endpoint address in case of HTTP redirection */
@@ -1039,6 +1040,7 @@ device_stm_start_scan (device *dev)
     params->y_res = y_resolution;
     params->src = dev->opt.src;
     params->colormode = dev->opt.colormode_real;
+    params->scanintent = dev->opt.scanintent;
     params->format = device_choose_format(dev, src);
 
     /* Dump parameters */
@@ -1050,6 +1052,8 @@ device_stm_start_scan (device *dev)
             id_colormode_sane_name(dev->opt.colormode_emul));
     log_trace(dev->log, "  colormode_real: %s",
             id_colormode_sane_name(params->colormode));
+    log_trace(dev->log, "  scanintent:     %s",
+            id_scanintent_sane_name(params->scanintent));
     log_trace(dev->log, "  tl_x:           %s mm",
             math_fmt_mm(dev->opt.tl_x, buf));
     log_trace(dev->log, "  tl_y:           %s mm",
