@@ -86,6 +86,18 @@ typedef struct {
     bool          pdf_a;
     bool          png;
     bool          dib;
+
+    /* Quirks */
+
+    /* Scanner doesn't handle sca:ImagesToTransfer if set to "0"
+     * (which means "scan until ADF is empty").
+     *
+     * This is the Ricoh Aficio MP 201 case.
+     *
+     * The workaround is to set sca:ImagesToTransfer to some big
+     * arbitrary number.
+     */
+    bool          quirk_broken_ImagesToTransfer;
 } proto_handler_wsd;
 
 /* Forward declarations */
@@ -147,10 +159,20 @@ wsd_devcaps_query (const proto_ctx *ctx)
     xml_wr_enter(xml, "soap:Body");
     xml_wr_enter(xml, "sca:GetScannerElementsRequest");
     xml_wr_enter(xml, "sca:RequestedElements");
-    //xml_wr_add_text(xml, "sca:Name", "sca:ScannerDescription");
+
+    /* sca:ScannerConfiguration response defines scanner capabilities,
+     * such as document formats support, available sources, avaliable
+     * resolutions, color modes etc.
+     */
     xml_wr_add_text(xml, "sca:Name", "sca:ScannerConfiguration");
-    //xml_wr_add_text(xml, "sca:Name", "sca:DefaultScanTicket");
-    //xml_wr_add_text(xml, "sca:Name", "sca:ScannerStatus");
+
+    /* These elements are only requested for logging, to provide some
+     * device information for troubleshooting purposes
+     */
+    xml_wr_add_text(xml, "sca:Name", "sca:ScannerDescription");
+    xml_wr_add_text(xml, "sca:Name", "sca:DefaultScanTicket");
+    xml_wr_add_text(xml, "sca:Name", "sca:ScannerStatus");
+
     xml_wr_leave(xml);
     xml_wr_leave(xml);
     xml_wr_leave(xml);
@@ -239,6 +261,49 @@ wsd_devcaps_parse_formats (proto_handler_wsd *wsd,
     return err;
 }
 
+/* Parse supported content types and map them to scan intents
+ */
+static error
+wsd_devcaps_parse_content_types (devcaps *caps, xml_rd *xml,
+        unsigned int *scanintents_out)
+{
+    error        err = NULL;
+    unsigned int level = xml_rd_depth(xml);
+    size_t       prefixlen = strlen(xml_rd_node_path(xml));
+    unsigned int scanintents = 0;
+
+    (void) caps;
+
+    /* Decode supported content types */
+    while (!xml_rd_end(xml)) {
+        const char *path = xml_rd_node_path(xml) + prefixlen;
+
+        if (!strcmp(path, "/scan:ContentTypeValue")) {
+            const char *v = xml_rd_node_value(xml);
+
+            if (!strcmp(v, "Auto")) {
+                scanintents |= 1 << ID_SCANINTENT_AUTO;
+            } else if (!strcmp(v, "Text")) {
+                scanintents |= 1 << ID_SCANINTENT_DOCUMENT;
+            } else if (!strcmp(v, "Photo")) {
+                scanintents |= 1 << ID_SCANINTENT_PHOTO;
+            } else if (!strcmp(v, "Halftone")) {
+                scanintents |= 1 << ID_SCANINTENT_HALFTONE;
+            } else if (!strcmp(v, "Mixed")) {
+                scanintents |= 1 << ID_SCANINTENT_TEXTANDGRAPHIC;
+            } else {
+                log_debug(NULL, "unknown content type: %s", v);
+            }
+        }
+
+        xml_rd_deep_next(xml, level);
+    }
+
+    *scanintents_out = scanintents;
+
+    return err;
+}
+
 /* Parse size
  */
 static error
@@ -387,18 +452,6 @@ wsd_devcaps_parse_source (devcaps *caps, xml_rd *xml, ID_SOURCE src_id)
         min_hei = tmp;
     }
 
-    /* Workaround for yet another Kyocera bug. This device doesn't
-     * honor scan region settings. I.e., it understands it,
-     * properly mirrors in DocumentFinalParameters, but completely
-     * ignores when generating the image.
-     *
-     * So we can't rely on device's ability to clip the image and
-     * must implement clipping in software. It can be enforced
-     * in our backend by the following two lines:
-     */
-    min_wid = max_wid;
-    min_hei = max_hei;
-
     /* Save min/max width and height */
     src->min_wid_px = min_wid;
     src->max_wid_px = max_wid;
@@ -432,6 +485,7 @@ wsd_devcaps_parse_configuration (proto_handler_wsd *wsd,
     size_t       prefixlen = strlen(xml_rd_node_path(xml));
     bool         adf = false, duplex = false;
     unsigned int formats = 0;
+    unsigned int scanintents = 0;
     int          i;
 
     /* Parse configuration */
@@ -440,6 +494,8 @@ wsd_devcaps_parse_configuration (proto_handler_wsd *wsd,
 
         if (!strcmp(path, "/scan:DeviceSettings/scan:FormatsSupported")) {
             err = wsd_devcaps_parse_formats(wsd, caps, xml, &formats);
+        } else if (!strcmp(path, "/scan:DeviceSettings/scan:ContentTypesSupported")) {
+            err = wsd_devcaps_parse_content_types(caps, xml, &scanintents);
         } else if (!strcmp(path, "/scan:Platen")) {
             err = wsd_devcaps_parse_source(caps, xml, ID_SOURCE_PLATEN);
         } else if (!strcmp(path, "/scan:ADF/scan:ADFFront")) {
@@ -467,23 +523,59 @@ wsd_devcaps_parse_configuration (proto_handler_wsd *wsd,
 
         if (src != NULL) {
             src->formats = formats;
+            src->scanintents = scanintents;
+
+            /* Note, as we can clip in software, we indicate
+             * minimal scan region size for SANE as 0x0. But
+             * maximal size is defined by hardware
+             */
             src->win_x_range_mm.min = src->win_y_range_mm.min = 0;
             src->win_x_range_mm.max = math_px2mm_res(src->max_wid_px, 1000);
             src->win_y_range_mm.max = math_px2mm_res(src->max_hei_px, 1000);
         }
     }
 
-    /* Note, WSD uses slightly unusual model: instead of providing
-     * source configurations for simplex and duplex modes, it provides
-     * source configuration for ADF front, which is required (when ADF
-     * is supported by device) and for ADF back, which is optional
+    /* Workaround for Brother MFC-9340CDW
      *
-     * So we assume, that ADF front applies to both simplex and duplex
-     * modes, while ADF back applies only to duplex mode
+     * This device reports ADFSupportsDuplex as 0, but returns both
+     * ADFFront and ADFBack elements.
      *
-     * So if duplex is supported, we either merge front and back
-     * configurations, if both are present, or simply copy front
-     * to back, if back is missed
+     * As a workaround, we assume that if both ADFFront and ADFBack are
+     * present (temporary saved under the ID_SOURCE_ADF_SIMPLEX and
+     * ID_SOURCE_ADF_DUPLEX slots), scanner supports duplex mode,
+     * regardless of the ADFSupportsDuplex value it returns
+     */
+    if (adf && caps->src[ID_SOURCE_ADF_DUPLEX] != NULL) {
+        duplex = true;
+    }
+
+    /* Please note that the standard model for SANE and for our implementation
+     * involves having two separate configurations for the duplex ADF: one for
+     * simplex mode and another for duplex mode. In duplex mode, it is assumed
+     * that the front and back page scanning will have the same
+     * characteristics.
+     *
+     * However, WSD employs a slightly different model. Instead of providing
+     * separate source configurations for simplex and duplex modes, it offers a
+     * source configuration for the ADF front, which is required when the ADF
+     * is supported by the device, and an optional configuration for the ADF
+     * back.
+     *
+     * According to the specification, the ADF back configuration is optional.
+     * If the scanner indicates duplex support (via the ADFSupportsDuplex) but
+     * does not provide a separate ADFBack element, the ADFBack should be
+     * assumed to be the same as ADFFront.
+     *
+     * During the decoding process, we temporarily store the ADF front
+     * information under the ID_SOURCE_ADF_SIMPLEX and the ADF back information
+     * under the ID_SOURCE_ADF_DUPLEX slots, and then make adjustments.
+     *
+     * When adjusting, we assume that the ADF front applies to both simplex and
+     * duplex modes, while the ADF back applies only to duplex mode.
+     *
+     * Therefore, if duplex is supported, we either merge the front and back
+     * configurations if both are present or simply copy the front
+     * configuration to the back if the back configuration is missing.
      */
     if (adf && duplex) {
         log_assert(NULL, caps->src[ID_SOURCE_ADF_SIMPLEX] != NULL);
@@ -502,6 +594,25 @@ wsd_devcaps_parse_configuration (proto_handler_wsd *wsd,
         caps->src[ID_SOURCE_ADF_DUPLEX] = NULL;
     }
 
+    /* Workaround for yet another Kyocera bug. This device doesn't
+     * honor scan region settings. I.e., it understands it,
+     * properly mirrors in DocumentFinalParameters, but completely
+     * ignores when generating the image.
+     *
+     * So we can't rely on device's ability to clip the image and
+     * must implement clipping in software. It can be enforced
+     * in our backend by setting minimum image size equal to
+     * max size.
+     */
+    for (i = 0; i < NUM_ID_SOURCE; i ++) {
+        devcaps_source *src = caps->src[i];
+
+        if (src != NULL) {
+            src->min_wid_px = src->max_wid_px;
+            src->min_hei_px = src->max_hei_px;
+        }
+    }
+
     /* Check that we have at least one source */
     ID_SOURCE id_src;
     bool      src_ok = false;
@@ -529,6 +640,11 @@ wsd_devcaps_parse (proto_handler_wsd *wsd,
     xml_rd *xml;
     bool   found_configuration = false;
 
+    /* Fill "constant" part of device capabilities */
+    caps->units = 1000;
+    caps->protocol = wsd->proto.name;
+    caps->justification_x = caps->justification_y = ID_JUSTIFICATION_UNKNOWN;
+
     /* Parse capabilities XML */
     err = xml_rd_begin(&xml, xml_text, xml_len, wsd_ns_rd);
     if (err != NULL) {
@@ -576,10 +692,12 @@ wsd_devcaps_decode (const proto_ctx *ctx, devcaps *caps)
     http_data         *data = http_query_get_response_data(ctx->query);
     error             err;
 
-    caps->units = 1000;
-    caps->protocol = ctx->proto->name;
-    caps->justification_x = caps->justification_y = ID_JUSTIFICATION_UNKNOWN;
+    /* Setup quirks */
+    if (!strcmp(ctx->devinfo->model, "RICOH Aficio MP 201")) {
+        wsd->quirk_broken_ImagesToTransfer = true;
+    }
 
+    /* Parse device capabilities response */
     err = wsd_devcaps_parse(wsd, caps, data->bytes, data->size);
 
     return err;
@@ -633,7 +751,7 @@ wsd_fault_check (const proto_ctx *ctx)
 /* Decode fault response
  */
 static proto_result
-wsd_fault_decode (const proto_ctx *ctx)
+wsd_fault_decode (const proto_ctx *ctx, bool cleanup)
 {
     proto_handler_wsd *wsd = (proto_handler_wsd*) ctx->proto;
     proto_result      result = {0};
@@ -643,7 +761,7 @@ wsd_fault_decode (const proto_ctx *ctx)
     /* Parse XML */
     result.err = xml_rd_begin(&xml, data->bytes, data->size, wsd_ns_rd);
     if (result.err != NULL) {
-        result.next = PROTO_OP_FINISH;
+        result.next = cleanup ? PROTO_OP_CLEANUP : PROTO_OP_FINISH;
         result.status = SANE_STATUS_IO_ERROR;
         return result;
     }
@@ -709,6 +827,7 @@ wsd_scan_query (const proto_ctx *ctx)
     xml_wr                  *xml = xml_wr_begin("soap:Envelope", wsd_ns_wr);
     const char              *source = NULL;
     const char              *colormode = NULL;
+    const char              *contenttype = NULL;
     const char              *format = NULL;
     static const char       *sides_simplex[] = {"sca:MediaFront", NULL};
     static const char       *sides_duplex[] = {"sca:MediaFront", "sca:MediaBack", NULL};
@@ -736,6 +855,18 @@ wsd_scan_query (const proto_ctx *ctx)
         log_internal_error(ctx->log);
     }
 
+    switch (params->scanintent) {
+    case ID_SCANINTENT_UNSET:          break;
+    case ID_SCANINTENT_AUTO:           contenttype = "Auto"; break;
+    case ID_SCANINTENT_DOCUMENT:       contenttype = "Text"; break;
+    case ID_SCANINTENT_PHOTO:          contenttype = "Photo"; break;
+    case ID_SCANINTENT_HALFTONE:       contenttype = "Halftone"; break;
+    case ID_SCANINTENT_TEXTANDGRAPHIC: contenttype = "Mixed"; break;
+
+    default:
+        log_internal_error(ctx->log);
+    }
+
     /* Create scan request */
     wsd_make_request_header(ctx, xml, WSD_ACTION_CREATE_SCAN_JOB);
 
@@ -804,7 +935,28 @@ wsd_scan_query (const proto_ctx *ctx)
     log_assert(ctx->log, format != NULL);
     xml_wr_add_text(xml, "sca:Format", format);
 
-    xml_wr_add_text(xml, "sca:ImagesToTransfer", "0");
+    /* WS-Scan specification says unspecified scan amount should be 0
+     * ( unknown amount, check for more ) and for Flatbed that is 1.
+     */
+    switch (params->src) {
+    case ID_SOURCE_PLATEN:
+        xml_wr_add_text(xml, "sca:ImagesToTransfer", "1");
+        break;
+    case ID_SOURCE_ADF_SIMPLEX:
+    case ID_SOURCE_ADF_DUPLEX:
+        if (wsd->quirk_broken_ImagesToTransfer) {
+            xml_wr_add_text(xml, "sca:ImagesToTransfer", "100");
+        } else {
+            xml_wr_add_text(xml, "sca:ImagesToTransfer", "0");
+        }
+        break;
+    default:
+        log_internal_error(ctx->log);
+    }
+
+    if (contenttype) {
+        xml_wr_add_text(xml, "sca:ContentType", contenttype);
+    }
 
     xml_wr_enter(xml, "sca:InputSize");
     xml_wr_enter(xml, "sca:InputMediaSize");
@@ -863,7 +1015,7 @@ wsd_scan_decode (const proto_ctx *ctx)
 
     /* Decode error, if any */
     if (wsd_fault_check(ctx)) {
-        return wsd_fault_decode(ctx);
+        return wsd_fault_decode(ctx, false);
     }
 
     /* Decode CreateScanJobResponse */
@@ -961,13 +1113,13 @@ wsd_load_decode (const proto_ctx *ctx)
 
     /* Check HTTP status */
     if (wsd_fault_check(ctx)) {
-        return wsd_fault_decode(ctx);
+        return wsd_fault_decode(ctx, true);
     }
 
     /* We expect multipart message with attached image */
     data = http_query_get_mp_response_data(ctx->query, 1);
     if (data == NULL) {
-        result.next = PROTO_OP_FINISH;
+        result.next = PROTO_OP_CLEANUP;
         result.err = ERROR("RetrieveImageRequest: invalid response");
         return result;
     }
@@ -1031,6 +1183,13 @@ wsd_status_decode (const proto_ctx *ctx)
             result.status = SANE_STATUS_NO_DOCS;
             return result;
         }
+
+        /* Ricoh Aficio MP 201 reports "ADF empty" status this strange way
+         */
+        if (!strcmp(wsd->fault_code, "ClientErrorJobIdNotFound")) {
+            result.status = SANE_STATUS_NO_DOCS;
+            return result;
+        }
     }
 
     /* Parse XML */
@@ -1150,6 +1309,18 @@ wsd_cancel_query (const proto_ctx *ctx)
     return wsd_http_post(ctx, xml_wr_finish_compact(xml));
 }
 
+/* Test interface: decode device capabilities
+ */
+static error
+wsd_test_decode_devcaps (proto_handler *proto,
+                         const void *xml_text, size_t xms_size,
+                         devcaps *caps)
+{
+    proto_handler_wsd *wsd = (proto_handler_wsd*) proto;
+
+    return wsd_devcaps_parse(wsd, caps, xml_text, xms_size);
+}
+
 /* proto_handler_wsd_new creates new WSD protocol handler
  */
 proto_handler*
@@ -1175,8 +1346,11 @@ proto_handler_wsd_new (void)
     wsd->proto.status_query = wsd_status_query;
     wsd->proto.status_decode = wsd_status_decode;
 
+    wsd->proto.cleanup_query = wsd_cancel_query;
     wsd->proto.cancel_query = wsd_cancel_query;
 
+    wsd->proto.test_decode_devcaps = wsd_test_decode_devcaps;
+
     return &wsd->proto;
 }
 
