@@ -11,6 +11,7 @@
 #include "airscan.h"
 
 #include <errno.h>
+#include <fnmatch.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
@@ -29,13 +30,8 @@
  */
 #define WSDD_RETRANSMIT_MIN     100     /* Min retransmit time */
 #define WSDD_RETRANSMIT_MAX     250     /* Max retransmit time */
-#define WSDD_DISCOVERY_TIME     2500    /* Overall discovery time */
-
-/* This delay is taken, if we have discovered, say, device's IPv6
- * addresses and have a strong suspicion that device has not yet
- * discovered IPv4 addresses as well
- */
-#define WSDD_PUBLISH_DELAY      1000
+#define WSDD_DISCOVERY_TIME     2500    /* Standard discovery time */
+#define WSDD_DISCOVERY_TIME_EX  5000    /* Extended discovery time */
 
 /* WS-Discovery stable endpoint path
  */
@@ -45,28 +41,35 @@
 /* wsdd_resolver represents a per-interface WSDD resolver
  */
 typedef struct {
-    int          fd;           /* File descriptor */
-    int          ifindex;      /* Interface index */
-    bool         ipv6;         /* We are on IPv6 */
-    eloop_fdpoll *fdpoll;      /* Socket fdpoll */
-    eloop_timer  *timer;       /* Retransmit timer */
-    uint32_t     total_time;   /* Total elapsed time */
-    ip_straddr   str_ifaddr;   /* Interface address */
-    ip_straddr   str_sockaddr; /* Per-interface socket address */
-    bool         initscan;     /* Initial scan in progress */
+    int           fd;             /* File descriptor */
+    int           ifindex;        /* Interface index */
+    netif_name    ifname;         /* Interface name, for logging */
+    bool          ipv6;           /* We are on IPv6 */
+    eloop_fdpoll  *fdpoll;        /* Socket fdpoll */
+    eloop_timer   *timer;         /* Retransmit timer */
+    uint32_t      time_limit;     /* Discovery time limit */
+    uint32_t      time_elapsed;   /* Elapsed time */
+    ip_straddr    str_ifaddr;     /* Interface address */
+    ip_straddr    str_sockaddr;   /* Per-interface socket address */
+    bool          initscan;       /* Initial scan in progress */
 } wsdd_resolver;
 
 /* wsdd_finding represents zeroconf_finding for WSD
  * device discovery
  */
 typedef struct {
-    zeroconf_finding  finding;        /* Base class */
-    const char        *address;       /* Device "address" in WS-SD sense */
-    ll_head           xaddrs;         /* List of wsdd_xaddr */
-    http_client       *http_client;   /* HTTP client */
-    ll_node           list_node;      /* In wsdd_finding_list */
-    eloop_timer       *publish_timer; /* WSDD_PUBLISH_DELAY timer */
-    bool              published;      /* This finding is published */
+    zeroconf_finding  finding;               /* Base class */
+    const char        *address;              /* Device WS-SD "address" */
+    ll_head           xaddrs;                /* List of wsdd_xaddr */
+    ll_head           xaddrs_unresolved;     /* wsdd_xaddr to be resolved */
+    zeroconf_endpoint *endpoints_unresolved; /* endpoints to be resolved */
+    mdns_resolver     *mdns_resolver;        /* MDNS resolver */
+    http_client       *http_client;          /* HTTP client */
+    ll_node           list_node;             /* In wsdd_finding_list */
+    eloop_timer       *publish_timer;        /* ZEROCONF_PUBLISH_DELAY timer */
+    bool              is_printer;            /* Device is printer */
+    bool              is_scanner;            /* Device is scanner */
+    bool              published;             /* This finding is published */
 } wsdd_finding;
 
 /* wsdd_xaddr represents device transport address
@@ -97,6 +100,12 @@ typedef struct {
 
 /* Forward declarations
  */
+static void
+wsdd_finding_add_xaddr (wsdd_finding *wsdd, wsdd_xaddr *xaddr);
+
+static void
+wsdd_finding_get_metadata (wsdd_finding *wsdd, int ifindex, wsdd_xaddr *xaddr);
+
 static void
 wsdd_message_free(wsdd_message *msg);
 
@@ -264,6 +273,9 @@ wsdd_finding_new (int ifindex, const char *address)
 
     wsdd->address = str_dup(address);
     ll_init(&wsdd->xaddrs);
+    ll_init(&wsdd->xaddrs_unresolved);
+
+    wsdd->mdns_resolver = mdns_resolver_new(ifindex);
     wsdd->http_client = http_client_new(wsdd_log, wsdd);
 
     return wsdd;
@@ -278,6 +290,9 @@ wsdd_finding_free (wsdd_finding *wsdd)
         zeroconf_finding_withdraw(&wsdd->finding);
     }
 
+    mdns_resolver_cancel(wsdd->mdns_resolver);
+    mdns_resolver_free(wsdd->mdns_resolver);
+
     http_client_cancel(wsdd->http_client);
     http_client_free(wsdd->http_client);
 
@@ -288,6 +303,8 @@ wsdd_finding_free (wsdd_finding *wsdd)
     zeroconf_endpoint_list_free(wsdd->finding.endpoints);
     mem_free((char*) wsdd->address);
     wsdd_xaddr_list_purge(&wsdd->xaddrs);
+    wsdd_xaddr_list_purge(&wsdd->xaddrs_unresolved);
+    zeroconf_endpoint_list_free(wsdd->endpoints_unresolved);
     ip_addrset_free(wsdd->finding.addrs);
     mem_free((char*) wsdd->finding.model);
     mem_free((char*) wsdd->finding.name);
@@ -318,7 +335,17 @@ wsdd_finding_publish (wsdd_finding *wsdd)
     zeroconf_finding_publish(&wsdd->finding);
 }
 
-/* WSDD_PUBLISH_DELAY timer callback
+/* wsdd_finding_has_pending_queries checks if wsdd_finding
+ * has pending MDNS or HTTP queries
+ */
+static bool
+wsdd_finding_has_pending_queries (wsdd_finding *wsdd)
+{
+    return mdns_resolver_has_pending(wsdd->mdns_resolver) ||
+           http_client_has_pending(wsdd->http_client);
+}
+
+/* ZEROCONF_PUBLISH_DELAY timer callback
  */
 static void
 wsdd_finding_publish_delay_timer_callback (void *data)
@@ -379,7 +406,7 @@ wsdd_finding_publish_delay (wsdd_finding *wsdd)
 
     if (delay) {
         if (wsdd->publish_timer == NULL) {
-            wsdd->publish_timer = eloop_timer_new(WSDD_PUBLISH_DELAY,
+            wsdd->publish_timer = eloop_timer_new(ZEROCONF_PUBLISH_DELAY,
                 wsdd_finding_publish_delay_timer_callback, wsdd);
         }
 
@@ -397,7 +424,7 @@ wsdd_finding_get (int ifindex, const char *address)
     ll_node      *node;
     wsdd_finding *wsdd;
 
-    /* Check for duplicates */
+    /* Lookup existent finding */
     for (LL_FOR_EACH(node, &wsdd_finding_list)) {
         wsdd = OUTER_STRUCT(node, wsdd_finding, list_node);
         if (wsdd->finding.ifindex == ifindex &&
@@ -423,10 +450,11 @@ wsdd_finding_by_address (ip_addr addr)
     wsdd_xaddr            *xaddr;
     const struct sockaddr *sockaddr;
 
-    /* Check for duplicates */
+    /* Roll over all findings */
     for (LL_FOR_EACH(node, &wsdd_finding_list)) {
         wsdd = OUTER_STRUCT(node, wsdd_finding, list_node);
 
+        /* Roll over all xaddrs */
         for (LL_FOR_EACH(node2, &wsdd->xaddrs)) {
             xaddr = OUTER_STRUCT(node2, wsdd_xaddr, list_node);
             sockaddr = http_uri_addr(xaddr->uri);
@@ -459,9 +487,111 @@ wsdd_finding_has_xaddr (wsdd_finding *wsdd, const wsdd_xaddr *xaddr)
         }
     }
 
+    for (LL_FOR_EACH(node, &wsdd->xaddrs_unresolved)) {
+        xaddr2 = OUTER_STRUCT(node, wsdd_xaddr, list_node);
+
+        if (http_uri_equal(xaddr->uri, xaddr2->uri)) {
+            return true;
+        }
+    }
+
     return false;
 }
 
+/* mdns_resolver callback for resolving non-literal
+ * hostnames in xaddrs
+ */
+static void
+wsdd_finding_mdns_resolver_xaddr_callback (const mdns_query *query)
+{
+    wsdd_finding     *wsdd = mdns_query_get_ptr(query);
+    const char       *host = mdns_query_get_name(query);
+    const ip_addrset *answer = mdns_query_get_answer(query);
+    size_t           count;
+    const ip_addr    *addrs = ip_addrset_addresses(answer, &count);
+    ll_node          *node;
+
+    for (LL_FOR_EACH(node, &wsdd->xaddrs_unresolved)) {
+        wsdd_xaddr *xaddr = OUTER_STRUCT(node, wsdd_xaddr, list_node);
+
+        if (http_uri_host_is(xaddr->uri, host)) {
+            size_t i;
+
+            for (i = 0; i < count; i ++) {
+                ip_addr    addr = addrs[i];
+                http_uri   *uri = http_uri_clone(xaddr->uri);
+                wsdd_xaddr *xaddr_resolved;
+
+                http_uri_set_host_addr(uri, addr);
+                xaddr_resolved = wsdd_xaddr_new(uri);
+
+                wsdd_finding_add_xaddr(wsdd, xaddr_resolved);
+            }
+        }
+    }
+}
+
+/* mdns_resolver callback for resolving non-literal
+ * hostnames in endpoints
+ */
+static void
+wsdd_finding_mdns_resolver_endpoint_callback (const mdns_query *query)
+{
+    wsdd_finding      *wsdd = mdns_query_get_ptr(query);
+    const char        *host = mdns_query_get_name(query);
+    const ip_addrset  *answer = mdns_query_get_answer(query);
+    size_t            count;
+    const ip_addr     *addrs = ip_addrset_addresses(answer, &count);
+    zeroconf_endpoint *ep;
+
+    for (ep = wsdd->endpoints_unresolved; ep != NULL; ep = ep->next) {
+        if (http_uri_host_is(ep->uri, host)) {
+            size_t i;
+
+            for (i = 0; i < count; i ++) {
+                ip_addr           addr = addrs[i];
+                http_uri          *uri = http_uri_clone(ep->uri);
+                zeroconf_endpoint *ep_resolved;
+
+                http_uri_set_host_addr(uri, addr);
+                ip_addrset_add(wsdd->finding.addrs, addr);
+
+                ep_resolved = zeroconf_endpoint_new(ID_PROTO_WSD, uri);
+                ep_resolved->next = wsdd->finding.endpoints;
+                wsdd->finding.endpoints = ep_resolved;
+            }
+        }
+    }
+}
+
+/* Add newly discovered xaddr and optionally initiate metadata query
+ * This function takes ownership over xaddr
+ */
+static void
+wsdd_finding_add_xaddr (wsdd_finding *wsdd, wsdd_xaddr *xaddr)
+{
+    /* xaddr already known? */
+    if (wsdd_finding_has_xaddr(wsdd, xaddr)) {
+        wsdd_xaddr_free(xaddr);
+        return;
+    }
+
+    if (http_uri_is_literal(xaddr->uri)) {
+        ll_push_end(&wsdd->xaddrs, &xaddr->list_node);
+        if (wsdd->is_scanner) {
+            wsdd_finding_get_metadata(wsdd, wsdd->finding.ifindex, xaddr);
+        }
+    } else {
+        ll_push_end(&wsdd->xaddrs_unresolved, &xaddr->list_node);
+        mdns_query_submit(
+            wsdd->mdns_resolver,
+            http_uri_get_host(xaddr->uri),
+            wsdd_finding_mdns_resolver_xaddr_callback,
+            wsdd
+        );
+    }
+}
+
 /* Delete wsdd_finding from the wsdd_finding_list
  */
 static void
@@ -508,17 +638,14 @@ wsdd_finding_list_purge (void)
  *
  * It ignores all endpoints except ScannerServiceType, extracts endpoint
  * URLs and prepends them to the wsdd->finding.endpoints
- *
- * Returns true if some endpoints were extracted, false otherwise
  */
-static bool
+static void
 wsdd_finding_parse_endpoints (wsdd_finding *wsdd, xml_rd *xml)
 {
     unsigned int      level = xml_rd_depth(xml);
     size_t            prefixlen = strlen(xml_rd_node_path(xml));
     bool              is_scanner = false;
     zeroconf_endpoint *endpoints = NULL;
-    bool              ok = false;
 
     while (!xml_rd_end(xml)) {
         const char *path = xml_rd_node_path(xml) + prefixlen;
@@ -548,24 +675,34 @@ wsdd_finding_parse_endpoints (wsdd_finding *wsdd, xml_rd *xml)
 
     if (!is_scanner) {
         zeroconf_endpoint_list_free(endpoints);
-        return false;
+        return;
     }
 
-    ok = endpoints != NULL;
-
     while (endpoints != NULL) {
         zeroconf_endpoint     *ep = endpoints;
         const struct sockaddr *addr = http_uri_addr(ep->uri);
 
+        endpoints = endpoints->next;
+
         if (addr != NULL) {
             ip_addrset_add(wsdd->finding.addrs, ip_addr_from_sockaddr(addr));
-        }
-        endpoints = endpoints->next;
-        ep->next = wsdd->finding.endpoints;
-        wsdd->finding.endpoints = ep;
-    }
+            ep->next = wsdd->finding.endpoints;
+            wsdd->finding.endpoints = ep;
+        } else if (!zeroconf_endpoint_list_contains(wsdd->endpoints_unresolved,
+                ep)) {
+            ep->next = wsdd->endpoints_unresolved;
+            wsdd->endpoints_unresolved = ep;
 
-    return ok;
+            mdns_query_submit(
+                wsdd->mdns_resolver,
+                http_uri_get_host(ep->uri),
+                wsdd_finding_mdns_resolver_endpoint_callback,
+                wsdd
+            );
+        } else {
+            zeroconf_endpoint_free_single(ep);
+        }
+    }
 }
 
 /* Get metadata callback
@@ -578,9 +715,6 @@ wsdd_finding_get_metadata_callback (void *ptr, http_query *q)
     http_data    *data;
     wsdd_finding *wsdd = ptr;
     char         *model = NULL, *manufacturer = NULL;
-    bool         ok = false;
-
-    (void) ptr;
 
     /* Check query status */
     err = http_query_error(q);
@@ -608,7 +742,7 @@ wsdd_finding_get_metadata_callback (void *ptr, http_query *q)
 
         if (!strcmp(path, "s:Envelope/s:Body/mex:Metadata/mex:MetadataSection"
                 "/devprof:Relationship/devprof:Hosted")) {
-            ok = wsdd_finding_parse_endpoints(wsdd, xml) || ok;
+            wsdd_finding_parse_endpoints(wsdd, xml);
         } else if (!strcmp(path, "s:Envelope/s:Body/mex:Metadata/mex:MetadataSection"
                 "/devprof:ThisModel/devprof:Manufacturer")) {
             if (manufacturer == NULL) {
@@ -644,15 +778,19 @@ wsdd_finding_get_metadata_callback (void *ptr, http_query *q)
         }
     }
 
-    /* Cancel cancel all unnecessary metadata requests.
+    /* Cancel all unnecessary metadata requests
      *
-     * Note, we consider request is unnecessary if:
-     *   * it has the same address family
-     *   * it belongs to the same network interface
+     * Depending on the device, metadata request may either return all
+     * endpoints, or metadata request sent to IPv4 address may return
+     * only IPv4 endpoints and visa versa
+     *
+     * But if we have both IPv4/IPv6 endpoints, seems waiting for completion
+     * of more metadata requests is not unnecessary. So lets cancel them,
+     * if any.
      */
-    if (ok) {
-        http_client_cancel_af_uintptr(wsdd->http_client,
-            http_uri_af(http_query_uri(q)), http_query_get_uintptr(q));
+    if (ip_addrset_has_af(wsdd->finding.addrs, AF_INET) &&
+        ip_addrset_has_af(wsdd->finding.addrs, AF_INET6)) {
+        http_client_cancel(wsdd->http_client);
     }
 
     /* Cleanup and exit */
@@ -661,7 +799,7 @@ DONE:
     mem_free(model);
     mem_free(manufacturer);
 
-    if (http_client_has_pending(wsdd->http_client) == 0) {
+    if (!wsdd_finding_has_pending_queries(wsdd)) {
         wsdd_finding_publish_delay(wsdd);
     }
 }
@@ -701,8 +839,14 @@ wsdd_message_parse_endpoint (wsdd_message *msg, xml_rd *xml)
 
         if (!strcmp(path, "/d:Types")) {
             val = xml_rd_node_value(xml);
-            msg->is_scanner = !!strstr(val, "ScanDeviceType");
-            msg->is_printer = !!strstr(val, "PrintDeviceType");
+
+            if (strstr(val, "ScanDeviceType")) {
+                msg->is_scanner = true;
+            }
+
+            if (strstr(val, "PrintDeviceType") != NULL) {
+                msg->is_printer = true;
+            }
         } else if (!strcmp(path, "/d:XAddrs")) {
             mem_free(xaddrs_text);
             xaddrs_text = str_dup(xml_rd_node_value(xml));
@@ -867,12 +1011,6 @@ wsdd_resolver_message_dispatch (wsdd_resolver *resolver,
     wsdd_xaddr   *xaddr;
     ll_node      *node;
 
-    /* Fixup ipv6 zones */
-    for (LL_FOR_EACH(node, &msg->xaddrs)) {
-        xaddr = OUTER_STRUCT(node, wsdd_xaddr, list_node);
-        http_uri_fix_ipv6_zone(xaddr->uri, resolver->ifindex);
-    }
-
     /* Write trace messages */
     log_trace(wsdd_log, "%s message received from %s:",
         wsdd_message_action_name(msg), from);
@@ -884,10 +1022,8 @@ wsdd_resolver_message_dispatch (wsdd_resolver *resolver,
         log_trace(wsdd_log, "  xaddr:      %s", http_uri_str(xaddr->uri));
     }
 
-    /* Handle the message */
-    switch (msg->action) {
-    case WSDD_ACTION_HELLO:
-    case WSDD_ACTION_PROBEMATCHES:
+    /* Handle some special cases */
+    if (msg->action != WSDD_ACTION_BYE) {
         /* Ignore devices that are neither scanner not printer */
         if (!(msg->is_scanner || msg->is_printer)) {
             log_trace(wsdd_log,
@@ -895,24 +1031,37 @@ wsdd_resolver_message_dispatch (wsdd_resolver *resolver,
             goto DONE;
         }
 
+        /* Ignore messages with no xaddrs */
+        if (ll_empty(&msg->xaddrs)) {
+            log_trace(wsdd_log,
+                "skipped: no xaddrs in the message");
+            goto DONE;
+        }
+    }
+
+    /* Fixup ipv6 zones */
+    for (LL_FOR_EACH(node, &msg->xaddrs)) {
+        xaddr = OUTER_STRUCT(node, wsdd_xaddr, list_node);
+        http_uri_fix_ipv6_zone(xaddr->uri, resolver->ifindex);
+    }
+
+    /* Handle the message */
+    switch (msg->action) {
+    case WSDD_ACTION_HELLO:
+    case WSDD_ACTION_PROBEMATCHES:
         /* Add a finding or get existent one */
         wsdd = wsdd_finding_get(resolver->ifindex, msg->address);
 
+        /* Update is_printer/is_scanner flags */
+        wsdd->is_printer = wsdd->is_printer || msg->is_printer;
+        wsdd->is_scanner = wsdd->is_scanner || msg->is_scanner;
+
         /* Import newly discovered xaddrs and initiate metadata
          * query
          */
         while ((node = ll_pop_beg(&msg->xaddrs)) != NULL) {
             xaddr = OUTER_STRUCT(node, wsdd_xaddr, list_node);
-
-            if (wsdd_finding_has_xaddr(wsdd, xaddr)) {
-                wsdd_xaddr_free(xaddr);
-                continue;
-            }
-
-            ll_push_end(&wsdd->xaddrs, &xaddr->list_node);
-            if (msg->is_scanner) {
-                wsdd_finding_get_metadata(wsdd, resolver->ifindex, xaddr);
-            }
+            wsdd_finding_add_xaddr(wsdd, xaddr);
         }
 
         /* If there is no pending metadata queries, it may mean
@@ -925,8 +1074,8 @@ wsdd_resolver_message_dispatch (wsdd_resolver *resolver,
          *
          * At this case we can publish device now
          */
-        if (http_client_has_pending(wsdd->http_client) == 0) {
-            if (msg->is_scanner) {
+        if (!wsdd_finding_has_pending_queries(wsdd)) {
+            if (wsdd->is_scanner) {
                 wsdd_finding_publish_delay(wsdd);
             } else {
                 wsdd_finding_publish(wsdd);
@@ -1030,28 +1179,84 @@ wsdd_resolver_read_callback (int fd, void *data, ELOOP_FDPOLL_MASK mask)
     }
 }
 
+/* Count discovered devices with model names matching the specified parent.
+ *
+ * Pattern is the glob-style expression, applied to the model name
+ * of discovered devices.
+ */
+static unsigned int
+wsdd_resolver_count_devices_by_model (wsdd_resolver *resolver, const char *pattern)
+{
+    unsigned int answer = 0;
+    ll_node      *node;
+    wsdd_finding *wsdd;
+
+    /* Lookup existent finding */
+    for (LL_FOR_EACH(node, &wsdd_finding_list)) {
+        wsdd = OUTER_STRUCT(node, wsdd_finding, list_node);
+        if (wsdd->finding.ifindex == resolver->ifindex &&
+            wsdd->finding.model != NULL &&
+            wsdd->published &&
+            fnmatch(pattern, wsdd->finding.model, 0) == 0) {
+                answer ++;
+        }
+    }
+
+    return answer;
+}
+
 /* Retransmit timer callback
  */
 static void
 wsdd_resolver_timer_callback (void *data)
 {
     wsdd_resolver *resolver = data;
+    const char    pattern[] = "Pantum*";
+
     resolver->timer = NULL;
 
-    if (resolver->total_time >= WSDD_DISCOVERY_TIME) {
-        eloop_fdpoll_free(resolver->fdpoll);
-        close(resolver->fd);
-        resolver->fdpoll = NULL;
-        resolver->fd = -1;
-        log_debug(wsdd_log, "%s: done discovery", resolver->str_ifaddr.text);
+    /* Resolver is about to use all its time limit.
+     * Should we use extended discovery time?
+     */
+    if (resolver->time_elapsed >= resolver->time_limit &&
+        resolver->time_limit < WSDD_DISCOVERY_TIME_EX) {
+        unsigned int cnt_mdns, cnt_wsdd;
 
-        if (resolver->initscan) {
-            resolver->initscan = false;
-            wsdd_initscan_count_dec();
+        cnt_mdns = mdns_device_count_by_model(resolver->ifindex, pattern);
+        cnt_wsdd = wsdd_resolver_count_devices_by_model(resolver, pattern);
+
+        if (cnt_wsdd < cnt_mdns) {
+            const char    *proto = resolver->ipv6 ? "ipv6" : "ipv4";
+
+            log_debug(wsdd_log, "%s@%s: \"%s\": MDNS/WSDD count: %d/%d",
+                proto, resolver->ifname.text,
+                pattern, cnt_mdns, cnt_wsdd);
+
+            log_debug(wsdd_log, "%s@%s: extending discovery time (%d->%d ms)",
+                proto, resolver->ifname.text,
+                resolver->time_limit, WSDD_DISCOVERY_TIME_EX);
+
+            resolver->time_limit = WSDD_DISCOVERY_TIME_EX;
         }
-    } else {
+    }
+
+    /* Time limit not reached yet? */
+    if (resolver->time_elapsed < resolver->time_limit) {
         wsdd_resolver_send_probe(resolver);
-    };
+        return;
+    }
+
+    /* Resolving is done, cleanup after that */
+    eloop_fdpoll_free(resolver->fdpoll);
+    close(resolver->fd);
+    resolver->fdpoll = NULL;
+    resolver->fd = -1;
+    log_debug(wsdd_log, "%s: done discovery", resolver->str_ifaddr.text);
+
+    if (resolver->initscan) {
+        resolver->initscan = false;
+        wsdd_initscan_count_dec();
+    }
 }
 
 /* Set retransmit timer
@@ -1063,13 +1268,28 @@ wsdd_resolver_timer_set (wsdd_resolver *resolver)
 
     log_assert(wsdd_log, resolver->timer == NULL);
 
-    if (resolver->total_time + WSDD_RETRANSMIT_MAX >= WSDD_DISCOVERY_TIME) {
-        t = WSDD_DISCOVERY_TIME - resolver->total_time;
+    if (resolver->time_elapsed >= resolver->time_limit) {
+        /* Should not happen, but just in case */
+        t = WSDD_RETRANSMIT_MIN;
     } else {
-        t = math_rand_range(WSDD_RETRANSMIT_MIN, WSDD_RETRANSMIT_MAX);
+        /* Choose random time between WSDD_RETRANSMIT_MIN
+         * and WSDD_RETRANSMIT_MAX, but don't exceed
+         * available limit, if possible (i.e., if it
+         * is not less that WSDD_RETRANSMIT_MIN)
+         */
+        t = resolver->time_limit - resolver->time_elapsed;
+        if (t < WSDD_RETRANSMIT_MIN) {
+            t = WSDD_RETRANSMIT_MIN;
+        } else {
+            if (t > WSDD_RETRANSMIT_MAX) {
+                t = WSDD_RETRANSMIT_MAX;
+            }
+
+            t = math_rand_range(WSDD_RETRANSMIT_MIN, t);
+        }
     }
 
-    resolver->total_time += t;
+    resolver->time_elapsed += t;
     resolver->timer = eloop_timer_new(t,
             wsdd_resolver_timer_callback, resolver);
 }
@@ -1117,11 +1337,12 @@ wsdd_resolver_new (const netif_addr *addr, bool initscan)
     int           af = addr->ipv6 ? AF_INET6 : AF_INET;
     const char    *af_name = addr->ipv6 ? "AF_INET6" : "AF_INET";
     int           rc;
-    static int    no = 0;
     uint16_t      port;
 
     /* Build resolver structure */
     resolver->ifindex = addr->ifindex;
+    resolver->ifname = addr->ifname;
+    resolver->time_limit = WSDD_DISCOVERY_TIME;
 
     /* Open a socket */
     resolver->ipv6 = addr->ipv6;
@@ -1146,10 +1367,6 @@ wsdd_resolver_new (const netif_addr *addr, bool initscan)
         if (rc < 0) {
             goto FAIL;
         }
-
-        /* Note: error is not a problem here */
-        (void) setsockopt(resolver->fd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
-                &no, sizeof(no));
     } else {
         rc = setsockopt(resolver->fd, IPPROTO_IP, IP_MULTICAST_IF,
                 &addr->ip.v4, sizeof(addr->ip.v4));
@@ -1165,10 +1382,6 @@ wsdd_resolver_new (const netif_addr *addr, bool initscan)
         if (rc < 0) {
             goto FAIL;
         }
-
-        /* Note: error is not a problem here */
-        (void) setsockopt(resolver->fd, IPPROTO_IP, IP_MULTICAST_LOOP,
-                &no, sizeof(no));
     }
 
     /* Bind the socket */
@@ -1555,12 +1768,8 @@ static void
 wsdd_netif_dump_addresses (const char *prefix, netif_addr *list)
 {
     while (list != NULL) {
-        char suffix[32] = "";
-
-        if (list->ipv6 && ip_is_linklocal(AF_INET6, &list->ip)) {
-            sprintf(suffix, "%%%d", list->ifindex);
-        }
-        log_debug(wsdd_log, "%s%s%s", prefix, list->straddr, suffix);
+        log_debug(wsdd_log, "%s%s@%s, ifindex=%d", prefix, list->straddr,
+                list->ifname.text, list->ifindex);
         list = list->next;
     }
 }
